//hashmap of all subscribed commands and their handle callback
var subscribedCommands = {}; //each entry is a handle callback with the command as key

//called by a subscriber, to handle the given command
var brokerHandle = function(command, cb) {
    //if the first argument is not a string, it is considered to be a commands object like {command : function(...)}
    if (typeof(command) !== 'string') {
        brokerHandleMulti(command);
        return;
    }
    
    //cannot handle the same command multiple times
    if (subscribedCommands[command]) throw "Duplicate Command Handle: Cannot handle the same command twice.";
    
    //add the callback to this command
    subscribedCommands[command] = cb;
    
    //return broker so that handle can be chained
    return this;
};

//same as brokerHandle, but allows multiple handles in one call
var brokerHandleMulti = function(commandsObj) {
    for (var command in commandsObj) {
        brokerHandle(command, commandsObj[command]);
    }
};

//called by a subscriber, to remove the handle to a given command
var brokerRemoveHandle = function(command) {
    delete subscribedCommands[command];
    return this; //return broker so that rem handle can be chained
};

//called by publisher, to execute the given command
var brokerExecute = function(command) {
    //if command is not a string it is a control flow object for multiple executions
    if(typeof(command) !== 'string') {
        brokerExecuteFlow.apply(this, arguments);
        return this; //return broker so that exec can be chained
    }
    
    //make sure that this command is being listened to; otherwise a registration order mistake has been made
    if (!subscribedCommands[command]) throw "Cannot execute given command as it does not exist yet.";
    
    //convert arguments obj to array and remove the command argument
    var args = Array.prototype.slice.call(arguments);
    args.shift();
    
    //call subscribed callback for this command with the given arguments
    subscribedCommands[command].apply(null, args);
    
    //return broker so that exec can be chained
    return this;
};

//called by publisher to execute multiple commands followed by a done callback
var brokerExecuteFlow = function(commandsObj, doneCallback) {
    var callbacksQueue = 0; //set queue of the amount of callbacks that still need to be fired
    //go through each command
    var commandsHandled = false; //not all commands have been handled yet
    for (var command in commandsObj) {
        var commandArgs = commandsObj[command];
        if (!commandArgs) { //no arguments are set (or to null)
            //no callback has to be recorded here, just execute it
            brokerExecute(command);
        }
        else if (typeof(commandArgs) === 'function') {
            callbacksQueue++;
            //it only has a callback, wrap it to detect execution
            var cbWrapper = function() {
                commandArgs.apply(this, arguments);
                callbacksQueue--;
                
                //check if it is time to call done callback (make sure that the others have executed too)
                if (commandsHandled && callbacksQueue == 0) doneCallback();
            };
            brokerExecute(command, cbWrapper);
        }
        else {
            //it's an array of arguments, last function is considered to be the callback (if present)
            //loop from end to start
            for (var i = commandArgs.length - 1; i >= 0; i--) {
                //if this is a function, it is considered to be the callback since it is the last one
                if (typeof(commandArgs[i]) === 'function') {
                    callbacksQueue++;
                    var origFunc = commandArgs[i];
                    var cbWrapper = function() {
                        origFunc.apply(this, arguments);
                        callbacksQueue--;
                        
                        //check if it is time to call done callback (make sure that the others have executed too)
                        if (commandsHandled && callbacksQueue == 0) doneCallback();
                    };
                    commandArgs[i] = cbWrapper;
                    break;
                }
            }
            //execute command, also when there is no callback (in that case queue will not have been updated either)
            commandArgs.unshift(command);
            brokerExecute.apply(this, commandArgs);
        }
    }
    commandsHandled = true; //all commands have now been handled
    if (callbacksQueue == 0) doneCallback();
};

//add browser support, can be accessed anywhere using eg. broker.exec by using the window property
if (typeof(window) !== 'undefined') {
    var exports = {};
	window.broker = exports;
}

//published methods
exports.handle = brokerHandle;
exports.exec = brokerExecute;
exports.removeHandle = brokerRemoveHandle;